#include "BitHeap.h"

//   5
// Sigma 32^i = (32^(5+1) - 1)/(32 - 1) = 1073741823/31 = 34636833
//  i=0

enum { Arity         = 32,
       Start         = 25,
       Width         = 5,
       Bitheap_size  = 34636833 };

static unsigned Bitheap[Bitheap_size] = { 0x00000000u };

static void * Min;
static void * Max;

static unsigned get(unsigned Value, unsigned Offset, unsigned Index)
  { unsigned bit = Value >> Offset;
    if (!(Bitheap[Index] & (1 << bit)))
      return 0;
    if (Offset > 0)
      return get(Value & (1 << Offset) - 1, Offset - Width, 1 + (Index << Width) + bit);
    else
      return (Bitheap[Index] & (1 << Value))? 1 : 0; }

static void set(unsigned Value, unsigned Offset, unsigned Index)
  { if (Offset > 0)
      { unsigned bit = Value >> Offset;
        Bitheap[Index] |= 1 << bit;
        set(Value & (1 << Offset) - 1, Offset - Width, 1 + (Index << Width) + bit); }
    else
      Bitheap[Index] |= 1 << Value; }

static void foreach(unsigned Depth, unsigned Index, void (* Proc)(void *))
  { unsigned bitnode = Bitheap[Index];
    if (bitnode)
      { Index = 1 + (Index << Width);
        for (unsigned bit = 0; bit < Arity; bit += 1)
          if (bitnode & (1 << bit))
            { unsigned index = Index + bit;
              if (Depth > 0)
                foreach(Depth - 1, index, Proc);
              else
                Proc(Min + index - Bitheap_size); }}}

void BitHeap_clear(void * First, void * Last)
  { Min = First;
    Max = Last;
    for (unsigned index = 0; index < Bitheap_size; index += 1)
      Bitheap[index] = 0; }

void BitHeap_set(void * Address)
  { set((unsigned)(Address - Min), Start, 0); }

void BitHeap_get(void * Address)
  { get((unsigned)(Address - Min), Start, 0); }

void BitHeap_foreach(void (* Proc)(void *))
  { foreach(Width, 0, Proc); }
