#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "Jonkers_Schorr_Waite(bitheap).h"

static const unsigned Size = 1000;

const ptr Null = 0;
const unsigned Max = 100000;

static cel memory[Size];

ptr Memory, First, Next, Free, Last;

unsigned is_raw(ptr P)
  { return (signed)P < 0; }

unsigned size(ptr P)
  { return abs((signed)P); }

unsigned is_chunk(ptr P)
  { return (long)P < (long)Next; };

unsigned is_left(ptr P)
  { return !((long)P - (long)Memory) / sizeof(cel) % 2; }

unsigned is_right(ptr P)
  { return ((long)P - (long)Memory) / sizeof(cel) % 2; }

ptr stretch(unsigned U)
  { return (ptr)((long)U - 1); }

static cel make_chunk(unsigned size)
  { cel C = (cel){ Next, p, u };
    *Next++ = (cel){ (ptr)(long)size, h, u };
    for ( ; size ; size--)
      *Next++ = (cel){ Null, a, u };
    return C; }

static cel car(cel cons)
  { assert(cons.T == p);
    return *cons.P; }

static cel cdr(cel cons)
  { assert(cons.T == p);
    return *(cons.P + 1); }

static void set_car(cel cons, cel car)
  { assert(cons.T == p);
    *cons.P = car; }

static void set_cdr(cel cons, cel cdr)
  { assert(cons.T == p);
    *(cons.P + 1) = cdr; }

static cel cons(cel car, cel cdr)
  { ptr pointer;
    if (Free == Null)
      { assert(First > Next + 2 + 1);
        pointer = First -= 2; }
    else
      { pointer = Free;
        Free = pointer->P; }
    *(pointer + 0) = car;
    *(pointer + 1) = cdr;
    return (cel){ pointer, p, u }; }

static void chunk_set(cel chunk, unsigned offset, cel value)
  { assert(chunk.T == p);
    *(chunk.P + offset + 1) = value; }

static cel chunk_get(cel chunk, unsigned offset)
  { assert(chunk.T == p);
    return *(chunk.P + offset + 1); }

static void show(char* Text)
  { printf("%s:\n", Text);
    printf("  Chunks:");
    for (ptr p = Memory; p < Next; p += 1)
      printf("\n   *0x%010lx = [ 0x%010lx , %d , %d ]", (long)p, (long)p->P, p->T, p->M);
    printf("\n");
    printf("  Pairs:");
    for (ptr p = First; p < Last; p += 1)
      printf("\n   *0x%010lx = [ 0x%010lx , %d , %d ]", (long)p, (long)p->P, p->T, p->M);
    printf("\n   Free = 0x%010lx", (long)Free);
    printf("\n\n"); }

int main(int argc, const char * argv[])
  { cel r1, r2, r3;
    cel A, B, C, D, E, F, G;
    Memory = memory;
    *Memory = (cel){ Null, a, u };
    Next = Memory + 1;
    Free = Null;
    Last = Memory + Size;
    if (is_right(Last))
      Last -= 1;
    First = Last;
    make_chunk(0);
    r1 = make_chunk(1);
    make_chunk(2);
    r2 = make_chunk(3);
    make_chunk(1);
    r3 = make_chunk(2);

    cel _11  = { (ptr)11, a, u };
    cel _12  = { (ptr)12, a, u };
    cel _13  = { (ptr)13, a, u };
    cel _14  = { (ptr)14, a, u };
    cel _15  = { (ptr)15, a, u };
    cel _21  = { (ptr)21, a, u };
    cel _22  = { (ptr)22, a, u };
    cel _31  = { (ptr)31, a, u };
    cel _32  = { (ptr)32, a, u };
    cel _33  = { (ptr)33, a, u };
    cel _34  = { (ptr)34, a, u };
    cel _35  = { (ptr)35, a, u };
    cel _41  = { (ptr)41, a, u };
    cel _42  = { (ptr)42, a, u };
    cel _43  = { (ptr)43, a, u };

    A = cons(_11, _12);
    B = cons(_13, cons(_14, _15));
    C = cons(_21, _22);
    D = cons(_31 , B);
    E = cons(_32, _33);
    F = cons(r2, cons(_34, r1));
    set_car(cdr(F), cdr(cdr(cdr(D))));
    G = cons(_35, cons(_41, cons(_42, _43)));

    chunk_set(r1, 0, r2);
    chunk_set(r2, 0, r1);
    chunk_set(r2, 1, F);
    chunk_set(r2, 2, r3);
    chunk_set(r3, 0, r2);
    chunk_set(r3, 1, r1);

    show("Init");
    Jonkers_Schorr_Waite_bitheap_mark_thread(r1, Memory + 1, Next);
    show("Mark thread");
    Jonkers_Schorr_Waite_bitheap_unthread();
    show("Unthread");
    Next = Jonkers_Schorr_Waite_bitheap_compact();
    show("Compact");
    Jonkers_Schorr_Waite_bitheap_link();
    show("Link");

    return 0; }
